/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.netDesignAlgorithm.fa;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import com.net2plan.interfaces.networkDesign.IAlgorithm;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.libraries.CandidatePathList;
import com.net2plan.libraries.SRGUtils;
import com.net2plan.utils.Constants.RoutingType;
import com.net2plan.utils.Constants.SearchType;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.IntUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Triple;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Given a set of nodes and links, with their capacities, an offered traffic, and a set of admissible paths for each demand (given by the k-shortest paths
 * in km for each demand, being k a user-defined parameter), this algorithm uses a tabu search metaheuristic to find for each demand a SRG-disjoint
 * 1+1 routes, so that the network congestion is minimized. For each demand, the candidate 1+1 pairs are computed as all the SRG disjoint pairs p1,p2, such that
 * p1 and p2 are admissible paths. Two solutions are considered neighbors if they differ in the routing of one demand (either primary path, backup path or both).
 * The number of iterations is a user-defined parameter. The tabu list size is a user-defined parameters, defined as a fraction of the number of demands.
 * If the current solution is not improved during a number of consecutive iterations (a user-defined parameter), the algorithm produces a randomization of
 * the current solution based to jump to a different solution of the search space. This randomization makes use of a long-term memory structure. This
 * structure counts for each 1+1 pair p, the number of times that introducing this path in a solution, meant a solution improvement.
 * With this, we intend to capture the paths that more often have produced improving results, and use this information to tune the randomization jumps.
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.2, September 2015
 */
public class FA_TS_minCongestion11 implements IAlgorithm
{
	private final static boolean DEBUG = false;
	
	private Set<Long> linkIds, demandIds;
	private Map<Long, Double> u_e, h_d;
	private CandidatePathList cpl;
	private Map<Long, List<Pair<Long, Long>>> candidate11PairsList_d;
	
	@Override
	public String executeAlgorithm(NetPlan netPlan, Map<String, String> algorithmParameters, Map<String, String> net2planParameters)
	{
		/* Basic checks */
		final int N = netPlan.getNumberOfNodes();
		final int D = netPlan.getNumberOfDemands();
		final int E = netPlan.getNumberOfLinks();
		if (N == 0 || E == 0 || D == 0) throw new Net2PlanException("This algorithm requires a topology with links and a demand set");

		linkIds = netPlan.getLinkIds();
		demandIds = netPlan.getDemandIds();
		u_e = netPlan.getLinkCapacityMap();
		h_d = netPlan.getDemandOfferedTrafficMap();

		/* Initialize some variables */
		final int k = Integer.parseInt(algorithmParameters.get("k")); /* number of loopless candidate paths per demand */
		final int ts_maxNumIt = Integer.parseInt(algorithmParameters.get("ts_maxNumIt"));
		final int ts_tabuListTenure = (int) Math.floor(D * Double.parseDouble(algorithmParameters.get("ts_tabuListTenureFraction")));
		final int ts_numItNonImprovingToRandomize = Integer.parseInt(algorithmParameters.get("ts_numItNonImprovingToRandomize"));
		final boolean setOneSRGPerBidirectionalLink = Boolean.parseBoolean(algorithmParameters.get("setOneSRGPerBidirectionalLink"));

		/* Sets the network SRGs, if needed */
		if (setOneSRGPerBidirectionalLink) SRGUtils.configureSRGs(netPlan, 365 * 24, 12, SRGUtils.SharedRiskModel.PER_BIDIRECTIONAL_LINK_BUNDLE, true);
		
		/* Sets the routing to source routing */
		netPlan.setRoutingType(RoutingType.SOURCE_ROUTING);
		
		/* Construct the candidate path list */
		cpl = new CandidatePathList(netPlan, netPlan.getLinkLengthInKmMap(), "K", Integer.toString(k));

		/* For each path, compute those of the same demand which are link disjoint */
		candidate11PairsList_d = computeCandidate11PairList(cpl, netPlan);

		/* Solution coding: index of the candidate 1+1 pair associated to this demand */
		Map<Long, Integer> current_solution = IntUtils.zeros(demandIds); /* initialize it so all demands use the first 1+1 pair in its 1+1 pair list */
		double current_cost = computeNetCongestion(current_solution);

		/* Initialize tabu list: list of last demands changed */
		LinkedList<Long> tabuList = new LinkedList<Long>();

		/* Initialize long-time memory structure: for each 1+1 pair, the number of times appeared in improving solutions */
		Map<Long, int[]> longTermMemory = new LinkedHashMap<Long, int[]>();
		for (long demandId : demandIds) longTermMemory.put(demandId, new int[candidate11PairsList_d.get(demandId).size()]);

		/* Main loop */
		Random rnd = new Random();
		Map<Long, Integer> best_solution = new LinkedHashMap<Long, Integer>(current_solution);
		double best_cost = current_cost;
		int numberOfIterationsWithoutImproving = 0;
		for (int numIt = 0; numIt < ts_maxNumIt; numIt++)
		{
			/* For debug purposes: print information of the iteration */
			if (DEBUG) System.out.println("Starting iteration : " + numIt + ". Best cost: " + best_cost + ". Current cost: " + current_cost + ". Current solution: " + current_solution);

			/* check if the best solution was not improved in last iterations. If so, move to a new solution.
			 * This new solution is a random variation of a solution created from the long term memory */
			if (numberOfIterationsWithoutImproving > ts_numItNonImprovingToRandomize)
			{
				current_solution = new LinkedHashMap<Long, Integer>();
				for (long demandId : demandIds)
				{
					boolean takeBestLongTermPath = (rnd.nextDouble() <= 0.8); /* 0.8 probability takes the best pair according to long term memory */
					if (takeBestLongTermPath)
					{
						int[] bestPairIds = IntUtils.maxIndexes(longTermMemory.get(demandId), SearchType.ALL);
						current_solution.put(demandId, bestPairIds[rnd.nextInt(bestPairIds.length)]); /* the best pair according to long term memory. If more than one with the same => choose randomly among them */
					}
					else
					{
						current_solution.put(demandId, rnd.nextInt(longTermMemory.get(demandId).length)); /* a path randomly chosen */
					}
				}
				
				current_cost = computeNetCongestion(current_solution);
				numberOfIterationsWithoutImproving = 0;
			}

			/* compute the neighborhood: change one demand */
			Map<Long, Integer> bestAcceptableNeighbor_solution = null; /* best neighbor solution found in this iteration */
			double bestAcceptableNeighbor_cost = Double.MAX_VALUE; /* cost of best neighbor solution found in this iteration */
			long bestAcceptableNeighbor_changingDemand = -1; /* the demand that changes in the neighbor respect to current solution: needed for long-term memory update */

			/* Iterate for each neighbor: a neighbor solution differs from current in the 1+1 pair used in ONE demand  */
			for (long neighborChangingDemand : demandIds)
			{
				int numCandidatePairs_thisDemand = candidate11PairsList_d.get(neighborChangingDemand).size();
				for (int neighborNewPathPairId = 0; neighborNewPathPairId < numCandidatePairs_thisDemand; neighborNewPathPairId++)
				{
					/* If the neighbor is the same as current solution, skip this iteration */
					if (neighborNewPathPairId == current_solution.get(neighborChangingDemand)) continue;

					/* create the neighbor solution, copying it in a new array */
					Map<Long, Integer> neighbor_solution = new LinkedHashMap<Long, Integer>(current_solution);
					neighbor_solution.put(neighborChangingDemand, neighborNewPathPairId);

					/* compute the cost of the neighbor solution */
					double neighbor_cost = computeNetCongestion(neighbor_solution);

					/* Update the long-term memory structure: if the solution is better than current solution, store it */
					if (neighbor_cost < current_cost) longTermMemory.get(neighborChangingDemand)[neighborNewPathPairId]++;

					/* Ambition criteria: if the cost is better than best solution so far: accept it as next potential solution (best solution, and best neighbor are updated)  */
					if (neighbor_cost < best_cost)
					{
						best_solution = new LinkedHashMap<Long, Integer>(neighbor_solution);
						best_cost = neighbor_cost;
						bestAcceptableNeighbor_solution = new LinkedHashMap<Long, Integer>(neighbor_solution);
						bestAcceptableNeighbor_cost = neighbor_cost;
						numberOfIterationsWithoutImproving = 0;
						
						continue;
					}

					/* check if the neighbor solution is tabu: if so, skip it */
					if (tabuList.contains(neighborChangingDemand)) continue;

					/* the neighbor solution is not tabu: check if it is the best neighbor found so far */
					if (neighbor_cost < bestAcceptableNeighbor_cost)
					{
						bestAcceptableNeighbor_cost = neighbor_cost;
						bestAcceptableNeighbor_solution = new LinkedHashMap<Long, Integer>(neighbor_solution);
						bestAcceptableNeighbor_changingDemand = neighborChangingDemand;
					}
				}
			}

			/* Check if the solution we make is an improving solution (update numberOfIterationsWithoutImproving) */
			if (bestAcceptableNeighbor_cost < current_cost) numberOfIterationsWithoutImproving = 0;
			else numberOfIterationsWithoutImproving++;

			/* Move to the best neighbor found: current solution is the best neighbor */

			/* If there are no neighbors => rare thing => force random variation of the solution */
			if (bestAcceptableNeighbor_solution == null)
			{
				numberOfIterationsWithoutImproving = ts_numItNonImprovingToRandomize + 1;
				continue;
			}

			current_solution = new LinkedHashMap<Long, Integer>(bestAcceptableNeighbor_solution);
			current_cost = bestAcceptableNeighbor_cost;

			/* Update the tabu list: add the new move to the list */
			tabuList.addLast(bestAcceptableNeighbor_changingDemand);

			/* Update the tabu list: if the tabu list is oversized, remove the first (oldest) element  */
			if (tabuList.size() > ts_tabuListTenure) tabuList.removeFirst();
		}

		/* Save the best solution found into the netPlan object */
		netPlan.removeAllRoutingInformation();
		netPlan.setRoutingType(RoutingType.SOURCE_ROUTING);
		for(long demandId : demandIds)
		{
			double h_d_thisDemand = h_d.get(demandId);
			
			final int bestPair = best_solution.get(demandId);
			final Pair<Long, Long> primaryAndBackupPath = candidate11PairsList_d.get(demandId).get(bestPair);
			
			final long routeId = netPlan.addRoute(demandId, h_d_thisDemand, cpl.getPathSequenceOfLinks(primaryAndBackupPath.getFirst()), null);
			final long segmentId = netPlan.addProtectionSegment(cpl.getPathSequenceOfLinks(primaryAndBackupPath.getSecond()), h_d_thisDemand, null);
			netPlan.addProtectionSegmentToRouteBackupSegmentList(segmentId, routeId);
		}

		return "Ok! Cost : " + best_cost;
	}

	@Override
	public String getDescription()
	{
		return "Given a set of nodes and links, with their capacities, an offered traffic, and a set of admissible paths for each demand (given "
				+ "by the k-shortest paths in km for each demand, being k a user-defined parameter), this algorithm uses a tabu search metaheuristic "
				+ "to find for each demand a link-disjoint 1+1 routes, so that the network congestion is minimized. For each demand, the candidate 1+1 pairs are "
				+ "computed as all the link disjoint pairs p1,p2, such that p1 and p2 are admissible paths. Two solutions are considered neighbors if they "
				+ "differ in the routing of one demand (either primary path, backup path or both). The number of iterations is a user-defined parameter. "
				+ "The tabu list size is a user-defined parameters, defined as a fraction of the number of demands. If the current solution is not improved "
				+ "during a number of consecutive iterations (a user-defined parameter), the algorithm produces a randomization of the current solution based "
				+ "to jump to a different solution of the search space. This randomization makes use of a long-term memory structure. This structure counts for "
				+ "each 1+1 pair p, the number of times that introducing this path in a solution, meant a solution improvement. With this, we intend to "
				+ "capture the paths that more often have produced improving results, and use this information to tune the randomization jumps.";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> algorithmParameters = new LinkedList<Triple<String, String, String>>();
		algorithmParameters.add(Triple.of("k", "10", "Number of candidate paths per demand"));
		algorithmParameters.add(Triple.of("ts_maxNumIt", "1000", "Number of iterations in the outer loop"));
		algorithmParameters.add(Triple.of("ts_tabuListTenureFraction", "0.1", "Size of the tabu list as a fraction of the number of demands"));
		algorithmParameters.add(Triple.of("ts_numItNonImprovingToRandomize", "10", "Number iterations non improving the best solution, which imply a large randomization"));
		algorithmParameters.add(Triple.of("setOneSRGPerBidirectionalLink", "#boolean# true", "If true, sets that the network has one SRG for each directly connected node pairs, with all the links between them. MTTF is one year, and MTTR 12 hours"));
		return algorithmParameters;
	}
	
	/* computes the congestion in the network (the objective function of the problem) */
	private double computeNetCongestion(Map<Long, Integer> solution_d)
	{
		Map<Long, Double> y_e = DoubleUtils.zeros(linkIds);
		
		for (long demandId : demandIds)
		{
			double h_d_thisDemand = h_d.get(demandId);
			final Pair<Long, Long> pathPair = candidate11PairsList_d.get(demandId).get(solution_d.get(demandId));
			
			for (long linkId : cpl.getPathSequenceOfLinks(pathPair.getFirst())) y_e.put(linkId, y_e.get(linkId) + h_d_thisDemand);
			for (long linkId : cpl.getPathSequenceOfLinks(pathPair.getSecond())) y_e.put(linkId, y_e.get(linkId) + h_d_thisDemand);
		}
		
		return DoubleUtils.maxValue(DoubleUtils.divide(y_e, u_e));
	}

	/* From a candidate path list, extract for each demand, the list of link-disjoint path pairs. There pairs are candidates for becoming primary-backup
	 * path pairs for the demand.  */
	private Map<Long, List<Pair<Long, Long>>> computeCandidate11PairList(CandidatePathList cpl , NetPlan np)
	{
		Map<Long, List<Pair<Long, Long>>> cpl11 = new LinkedHashMap<Long, List<Pair<Long, Long>>>();

		for (long demandId : demandIds)
		{
			cpl11.put(demandId, new ArrayList<Pair<Long, Long>>());
			Set<Long> pathIds = cpl.getDemandPaths(demandId);
			
			Iterator<Long> p1_it = pathIds.iterator();
			while(p1_it.hasNext())
			{
				final long p1 = p1_it.next();
				final List<Long> seqLinks_1 = cpl.getPathSequenceOfLinks(p1);
				final Set<Long> srgIds_p1 = getSRGsPath (seqLinks_1 , np); 
				
				Iterator<Long> p2_it = pathIds.iterator();
				while(p2_it.hasNext())
				{
					final long p2 = p2_it.next();
					if (p1 == p2) continue;
					
					final List<Long> seqLinks_2 = cpl.getPathSequenceOfLinks(p2);
					final Set<Long> commonSRGs = getSRGsPath (seqLinks_2 , np); commonSRGs.retainAll (srgIds_p1);
					
					if (commonSRGs.isEmpty()) /* no SRG in common => a valid 1+1 path */
						cpl11.get(demandId).add(Pair.of(p1, p2));
				}
			}

			if (cpl11.get(demandId).isEmpty()) throw new Net2PlanException("Demand " + demandId + " has no two link-disjoint paths");
		}

		return cpl11;
	}
	
	/* Returns the set of SRGs that affect the links or nodes of a path */
	private Set<Long> getSRGsPath (List<Long> seqLinks , NetPlan np)
	{
		Set<Long> res = new LinkedHashSet<Long> ();
		for (long e : seqLinks)
		{
			final long a_e = np.getLinkOriginNode(e);
			final long b_e = np.getLinkDestinationNode(e);
			res.addAll(np.getLinkSRGs(e));
			res.addAll(np.getNodeSRGs(a_e));
			res.addAll(np.getNodeSRGs(b_e));
		}
		return res;
	}
}